import sys
import unittest
import socket
from threading import Thread
import json
import ssl
from time import sleep
from http.server import BaseHTTPRequestHandler, HTTPServer

# add library output path
sys.path.append('swig')

from decision_service import *

def get_free_port():
	s = socket.socket(socket.AF_INET, type=socket.SOCK_STREAM)
	s.bind(('localhost', 0))
	address, port = s.getsockname()
	s.close()
	return port

class MockHandler(BaseHTTPRequestHandler):

	def __init__(self, mock_server, *args):
		self.mock_server = mock_server
		BaseHTTPRequestHandler.__init__(self, *args)

	def do_GET_config(self):
		# Process an HTTP GET request and return a response with an HTTP 200 status.
		self.send_response(200)
		self.send_header('Content-Type', 'application/json; charset=utf-8')
		self.end_headers()

		response_content = json.dumps({"ModelBlobUri": "http://localhost/model",
			"EventHubInteractionConnectionString": "conn1",
			"EventHubObservationConnectionString": "conn2"
		})
		self.wfile.write(response_content.encode('utf-8'))

		self.mock_server.get = self.mock_server.get + 1

	def do_GET(self):
		if self.path == "/config.json":
			return self.do_GET_config()

		self.send_response(404)
		self.end_headers()

	def do_POST(self):
		self.mock_server.post = self.mock_server.post + 1

		response_code = 404
		if (len(self.mock_server.post_status_codes) == 0):
			# default behavior
			# - respond with 201 (created) for known EventHubs
			# - respond with 404 otherwise
			if self.path.startswith("/interaction") or self.path.startswith("/observation"):
				response_code = 201
		else:
			response_code = self.mock_server.post_status_codes.pop(0)

		content_len = int(self.headers.get('content-length', 0))
		post_body = self.rfile.read(content_len).decode('utf-8')

		self.mock_server.posts.append({'path': self.path,
			'response_code': response_code,
			'body': post_body,
			'content-type': self.headers.get('content-type')})

		self.send_response(response_code)

		self.end_headers()

	# disable logging to stdout
	def log_message(self, format, *args):
		return

class MockServer:
	def __init__(self, https_enabled = True, post_status_codes = []):
		self.https_enabled = https_enabled
		self.post_status_codes = post_status_codes

	def __enter__(self):
		self.get = 0
		self.post = 0
		self.posts = []

		# Configure mock server.
		self.mock_server_port = get_free_port()

		def handler(*args):
			MockHandler(self, *args)

		self.mock_server = HTTPServer(('localhost', self.mock_server_port), handler)
		proto = 'http'
		if self.https_enabled:
			self.mock_server.socket = ssl.wrap_socket(self.mock_server.socket,
				server_side=True,
				certfile='unittest.pem')
			proto = 'https'

		# Start running mock server in a separate thread.
		# Daemon threads automatically shut down when the main process exits.
		self.mock_server_thread = Thread(target=self.mock_server.serve_forever)
		self.mock_server_thread.setDaemon(True)
		self.mock_server_thread.start()

		self.base_url = '{proto}://localhost:{port}/'.format(proto=proto, port=self.mock_server_port)

		return self

	def __exit__(self, exception_type, exception_value, trackback):
		self.mock_server.server_close()

class TestDecisionServiceConfiguration(unittest.TestCase):
	def test_download(self):
		# can't easily disable cert validation w/o introducing yet another method (optional params not supported)
		with MockServer(https_enabled = False) as server:
			config = DecisionServiceConfiguration_Download(server.base_url + 'config.json')

			self.assertEqual(config.model_url, 'http://localhost/model')
			self.assertEqual(config.eventhub_interaction_connection_string, 'conn1')
			self.assertEqual(config.eventhub_observation_connection_string, 'conn2')

			# make sure it's called once'
			self.assertEqual(server.get, 1, 'HTTP server not called')

	def test_not_found(self):
		with MockServer() as server:
			with self.assertRaises(SystemError):
				DecisionServiceConfiguration_Download(server.base_url + 'notfound')

class TestDecisionServiceListener(DecisionServiceListener):
	def __init__(self):
		self.messages = []
		DecisionServiceListener.__init__(self)

	def log(self, level, msg):
		self.messages.append({'level':level, 'msg': msg})

class TestDecisionServiceClient(unittest.TestCase):
	def __init__(self, *args):
		self.key = "SharedAccessKeyName=RootManageSharedAccessKey;SharedAccessKey=0t45SYSay1Rm8U4riSfgaaLIIKWPMzqJILmxpAQdOHg=" 
		self.config = DecisionServiceConfiguration()

		self.config.batching_timeout_in_milliseconds = 100
		self.config.certificate_validation_enabled = False

		unittest.TestCase.__init__(self, *args)

	def test_cert_validation_fails(self):
		with MockServer() as server:
			self.config.eventhub_interaction_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=interaction" % (server.mock_server_port, self.key)
			self.config.eventhub_observation_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=observation" % (server.mock_server_port, self.key)
			# self.listener = 
			# C++ owns the object by default. Read http://www.swig.org/Doc3.0/Python.html#Python_nn35 on how to reverse
			# TODO: I don't understand how this work...
			listener = TestDecisionServiceListener()
			self.config.listener = listener

			# make sure it's enabled and we'll internally throw
			self.config.certificate_validation_enabled = True

			client = DecisionServiceClient(self.config)
			ranking = client.rank('{"a":2}', '', [1,2,3])

			sleep(0.5)

			self.assertEqual(len(server.posts), 0, "requests should not hit the server as cert validation should fail")
			self.assertEqual(listener.messages, [{'level':1, 'msg':"Failed to upload event: 'Error in SSL handshake'"}])

	def test_rank_event_id_generated(self):
		with MockServer() as server:
			self.config.eventhub_interaction_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=interaction" % (server.mock_server_port, self.key)
			self.config.eventhub_observation_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=observation" % (server.mock_server_port, self.key)
			# self.config.set_listener(TestDecisionServiceListener())

			client = DecisionServiceClient(self.config)
			ranking = client.rank('{"a":2}', '', [1,2,3])

			sleep(0.5)
						
			self.assertEqual(len(server.posts), 1)
			self.assertEqual(server.posts[0]['path'], '/interaction/messages?timeout=60&api-version=2014-01')
			self.assertEqual(server.posts[0]['response_code'], 201)
			self.assertEqual(server.posts[0]['content-type'], 'application/atom+xml;type=entry;charset=utf-8')
			self.assertEqual(server.posts[0]['body'], '{"Version":"1","EventId":"' + ranking.event_id + '","a":[1,0],"c":{{"a":2}},"p":[0.8,0.2],"VWState":{"m":"m1"}}')

	def test_rank1(self):
		with MockServer() as server:
			self.config.eventhub_interaction_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=interaction" % (server.mock_server_port, self.key)
			self.config.eventhub_observation_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=observation" % (server.mock_server_port, self.key)
			# self.config.set_listener(TestDecisionServiceListener())

			client = DecisionServiceClient(self.config)
			ranking = client.rank('{"a":2}', 'abc', [1,2,3])

			sleep(0.5)
						
			self.assertEqual(len(server.posts), 1)
			self.assertEqual(server.posts[0]['path'], '/interaction/messages?timeout=60&api-version=2014-01')
			self.assertEqual(server.posts[0]['response_code'], 201)
			self.assertEqual(server.posts[0]['content-type'], 'application/atom+xml;type=entry;charset=utf-8')
			self.assertEqual(server.posts[0]['body'], '{"Version":"1","EventId":"abc","a":[1,0],"c":{{"a":2}},"p":[0.8,0.2],"VWState":{"m":"m1"}}')

	def test_update_model(self):
		with MockServer() as server:
			self.config.eventhub_interaction_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=interaction" % (server.mock_server_port, self.key)
			self.config.eventhub_observation_connection_string = "Endpoint=sb://localhost:%d/;%s;EntityPath=observation" % (server.mock_server_port, self.key)
			listener = TestDecisionServiceListener()
			self.config.listener = listener
			self.config.log_level = 4

			client = DecisionServiceClient(self.config)

			buf = bytearray(b'foo')
			client.update_model(buf)

			self.assertEqual(listener.messages, [{'level':1, 'msg':'update_model(len=3)'}])

if __name__ == '__main__':
	unittest.main()



import unittest
import sys

def list_tests_from(path):
	loader = unittest.TestLoader()
	suite = loader.discover(path)
	for t1 in suite:
		tests = t1._tests
		if len(tests):
			for t1 in tests:
				for t2 in t1._tests:
					tn = t2.__str__().split()
					print(path + "." + tn[1][1:-1] + "." + tn[0])

if __name__ == "__main__":
	for arg in sys.argv[1:]:
		list_tests_from(arg)



import socket
import sys
import argparse

#readline is under GPL licensing
#import readline

def recvall(s, n):
    buf=s.recv(n)
    ret=len(buf)
    while ret>0 and len(buf)<n:
        if buf[-1]=='\n': break
        tmp=s.recv(n)
        ret=len(tmp)
        buf=buf+tmp
    return buf

def _find_getch():
    try:
        import Carbon    # try OS/X
        Carbon.Evt       # unix doesn't have this
        def _getch():
            if Carbon.Evt.EventAvail(0x0008)[0]==0: # 0x0008 is the keyDownMask
                return ''
            else:
                (what,msg,when,where,mod)=Carbon.Evt.GetNextEvent(0x0008)[1]
                ch = chr(msg & 0x000000FF)
                if ord(ch) == 3:
                    raise Exception("control-c")
                return ch
        return _getch()
    except:
        pass

    try:
        import msvcrt   # try windows
        def _getch():
            ch = msvcrt.getch
            if ord(ch) == 3:
                raise Exception("control-c")
            return ch
    except:
        pass
    
    # POSIX system. Create and return a getch that manipulates the tty.
    try:
        import sys, tty, termios
        def _getch():
            fd = sys.stdin.fileno()
            old_settings = termios.tcgetattr(fd)
            try:
                tty.setraw(fd)
                ch = sys.stdin.read(1)
            finally:
                termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
                if ord(ch) == 3:
                    raise Exception("control-c")
                return ch

        return _getch
    except:
        pass

    return lambda: raw_input('')
        
def get_label(example,minus1,i,tag,pred, input_fn=raw_input):
    print '\nrequest for example %d: tag="%s", prediction=%g: %s'%(i,tag,pred,example)
    while True:
        label=input_fn('Provide? [0/1/skip]: ')
        if label == '1':
            break
        if label == '0':
            if minus1: label='-1'
            break
        if 0 < len(label) <= len('skip') and 'skip'.startswith(label):
            label = None
            break
    return label

parser = argparse.ArgumentParser(description="interact with VW in active learning mode")
parser.add_argument("-s","--seed", help="seed labeled dataset")
parser.add_argument("-o","--output", help="output file")
parser.add_argument("-u","--human", help="human readable examples in this file (same # of lines as unlabeled); use with -v")
parser.add_argument("-v","--verbose", action="store_true", help="show example (in addition to tag)")
parser.add_argument("-m","--minus1", action="store_true", help="interpret 0 as -1")
parser.add_argument("-k","--keypress", action="store_true", help="don't require 'Enter' after keypresses")
parser.add_argument("host", help="the machine VW is running on")
parser.add_argument("port", type=int, help="the port VW is listening on")
parser.add_argument("unlabeled_dataset", help="file with unlabeled data")
args = parser.parse_args()

input_fn = raw_input
if args.keypress:
    getch = _find_getch()
    def raw_input_keypress(str):
        print str,
        return getch()
    input_fn = raw_input_keypress

seed=None
if args.seed is not None:
    try:
        seed=open(args.seed,'r')
    except:
        print 'Warning: could not read from %s'%args.seed
        pass

try:
    unlabeled=open(args.unlabeled_dataset,'r')
except:
    print 'Error: could not read from %s'%args.unlabeled_dataset
    exit(1)

human=None
if args.human is not None:
    try:
        human = open(args.human, 'r')
    except:
        print 'Warning: could not read human examples from %s' % args.human
        pass
    
output=None
if args.output is not None:
    try:
        output=open(args.output,'w')
    except:
        print 'Warning: could not write to %s'%args.output
        pass


# Create a socket
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

queries=0
try:
    # Connect to server and perform handshake

    print 'connecting to %s:%d ...'%(args.host,args.port)
    sock.connect((args.host, args.port))
    print 'done'
    # Send seed dataset
    if seed:
        print 'seeding vw ...'
        for line in seed:
            sock.sendall(line)
            recvall(sock, 256)
        print 'done'
    # Send unlabeled dataset
    print 'sending unlabeled examples ...'
    for i,line in enumerate(unlabeled):
        humanString = line
        if human is not None:
            try:
                humanString = human.next()
            except:
                print 'Warning: out of lines in human data, reverting to vw strings'
                humanString = None
        sock.sendall(line)
        print 'sending unlabeled '+repr(line[:20])
        response=recvall(sock, 256)
        #print 'unlabeled response '+repr(response)
        responselist=response.split(' ')
        if len(responselist)==2:
            #VW does not care about this label
            continue
        prediction,tag,importance=responselist
        try:
            imp=float(importance)
        except:
            continue
        queries+=1
        label=get_label(humanString if args.verbose else '\n', args.minus1, i, tag, float(prediction), input_fn)
        if label is None:
            continue
        front,rest=line.split('|',1)
        if tag=='':
            tag="'empty"
        labeled_example=label+' '+"%g"%imp+' '+tag+' |'+rest
        #print 'sending labeled '+repr(labeled_example[:20])
        sock.sendall(labeled_example)
        if output:
            output.write(labeled_example)
            output.flush()
        recvall(sock, 256)
finally:
    sock.close()
    unlabeled.close()
    if output:
        output.close()
    if seed:
        seed.close()
    if human:
        human.close()



# -*- coding: utf-8 -*-
"""Vowpal Wabbit setup module """

import platform
import subprocess
import sys
from codecs import open
from ctypes.util import find_library
from distutils.command.clean import clean as _clean
from os import environ, makedirs, path, remove, walk
from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.test import test as _test
from setuptools.command.install_lib import install_lib as _install_lib
from shutil import copy, copytree, rmtree, ignore_patterns

system = platform.system()
here = path.abspath(path.dirname(__file__))
pylibvw = Extension('pylibvw', sources=['python/pylibvw.cc'])

def prep():
    """Prepare source directories for building extension """

    rmtree(path.join(here, 'src'), ignore_errors=True)

    subprocess.check_call(['git', 'checkout-index', '--all', '--prefix', 'python/src/'], cwd=path.join(here, '..'))

    # some pruning
    for d in ['cs', 'java', 'test', 'demo', 'logo_assets']:
        rmtree(path.join(here, 'src', d), ignore_errors=False)

class Clean(_clean):
    """Clean up after building python package directories """
    def run(self):
        try:
            remove('.coverage')
        except OSError:
            pass

        for d in ['.cache', '.tox', 'src', 'dist', 'build', 'vowpalwabbit.egg-info']:
            rmtree(path.join(here, d), ignore_errors=True)

        _clean.run(self)

class Sdist(_sdist):
    def run(self):
        # try to run prep if needed
        prep()
        _sdist.run(self)

class VWBuildExt(_build_ext):
    """Build pylibvw.so and install it as a python extension """
    def build_extension(self, ext):
	# moving the binary into install folder to make allow for multiple versions
        target_dir = path.join(here, path.dirname(self.get_ext_fullpath(ext.name)), 'vowpalwabbit')
        if not path.isdir(target_dir):
            makedirs(target_dir)

        if system == 'Windows':
            if sys.version_info[0] == 2 and sys.version_info[1] == 7:
               copy(path.join(here, 'bin', 'pyvw27.dll'), self.get_ext_fullpath(ext.name))
            elif sys.version_info[0] == 3 and sys.version_info[1] == 5:
               copy(path.join(here, 'bin', 'pyvw35.dll'), self.get_ext_fullpath(ext.name))
            elif sys.version_info[0] == 3 and sys.version_info[1] == 6:
               copy(path.join(here, 'bin', 'pyvw36.dll'), self.get_ext_fullpath(ext.name))
            else:
               raise Exception('Pre-built vw/python library for Windows is not supported for this python version')
        else:
            cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + target_dir,
                          '-DPIP_INSTALL=true']

            subprocess.check_call(['cmake', '.'] + cmake_args, cwd=path.join(here, 'src'))
            subprocess.check_call(['cmake', '--build', '.', '--', '-j'], cwd=path.join(here, 'src'))

class InstallLib(_install_lib):
    def build(self):
       _install_lib.build(self)
       if system == 'Windows':
           copy(path.join(here, 'bin', 'zlib.dll'), path.join(self.build_dir, 'zlib.dll'))

class Tox(_test):
    """ Run tox tests with 'python setup.py test' """
    tox_args = None
    test_args = None
    test_suite = None

    user_options = [('tox-args=', 'a', "Arguments to pass to tox")]

    def initialize_options(self):
        _test.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        _test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        import shlex
        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        errno = tox.cmdline(args=args)
        sys.exit(errno)


# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='vowpalwabbitds',
    version='8.5.0',
    description='Vowpal Wabbit Python package',
    long_description=long_description,
    url='https://github.com/JohnLangford/vowpal_wabbit',
    author='Scott Graham',
    author_email='scott.d.graham@gmail.com',
    license='BSD 3-Clause License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='fast machine learning online classification regression',
    packages=find_packages(exclude=['examples', 'src', 'tests']),
    platforms='any',
    zip_safe=False,
    include_package_data=True,
    ext_modules=[pylibvw],
    cmdclass={
        'build_ext': VWBuildExt,
        'clean': Clean,
        'sdist': Sdist,
        'test': Tox,
        'install_lib': InstallLib
    },
    # tox.ini handles additional test dependencies
    tests_require=['tox']
)
